#!/bin/sh -eu
#! -*- coding:utf-8; -*-

# Copyright (C) 2014 Moritz Klammler <moritz@klammler.eu>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

set -eu

envvars='
AWK
BASENAME
BIB
BIBFLAGS
DIRNAME
LN_S
MKDIR_P
SED
TEX
TEXFLAGS
abs_builddir
abs_srcdir
'
PROGRAM='run-latex.sh'
VERSION='1.1'
USAGE="$PROGRAM [OPTION...] [--] TARGET.pdf DEPENDENCY..."
helptext="usage: $USAGE

Produce TARGET.pdf by running LaTeX on TARGET.tex making sure that all
DEPENDENCIEs are available in the build directory.  VPATH resolution will be
done on the dependencies.  The file name extension of TARGET need not be
'.pdf', it can also be '.dvi', '.ps' or whatever.


The following OPTIONs are available:

--help      show this text and exit
--version   show version information and exit
--verbose   be more verbose (can be repeated)


The following environment variables *must* be provided:

$(for v in $envvars; do printf '  %s\n' $v; done; unset v)


Exits Status:

On successful operation the exit status is 0, on error, it is 1 and if invalid
arguments are passed, it will be 2."
versiontext="$PROGRAM $VERSION
Copyright (C) 2013 Moritz Klammler <moritz@klammler.eu>
This is free software; see the source for copying conditions.  There is NO
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE."

TEXLIMIT=${TEXLIMIT:-8}
verbosity=0
target=
dependencies=

## usage: error MESSAGE...
##
## Print the error MESSAGEs and exit with status 1.
##
error()
{
    echo "$PROGRAM: error: $@" >&2
    exit 1
}

## usage: uerror MESSAGE...
##
## Print the error MESSAGEs and the usage synopsis, then exit with status 2.
##
uerror()
{
    echo "$PROGRAM: error: $@" >&2
    echo "usage: $USAGE" >&2
    exit 2
}

## usage: info [FLAGS] LEVEL MESSAGE...
##
## Print MESSAGEs if and only if LEVEL is less or equal than the verbosity
## level.  If specified, FLAGS (eg '-e') will be passed on to echo.
info()
{
    flags=''
    case "$1" in
        -*)
            flags="$1"
            shift
            ;;
    esac
    if test $verbosity -ge $1
    then
        shift
        echo $flags "$PROGRAM: info: $@" >&2
    fi
    unset flags
}

onlyargs=0
for arg in "$@"
do
    if test $onlyargs -eq 0
    then
        case "$arg" in
            --help)
                echo "$helptext"
                exit 0
                ;;
            --version)
                echo "$versiontext"
                exit 0
                ;;
            --verbose)
                verbosity=$(($verbosity + 1))
                ;;
            --)
                onlyargs=1
                ;;
            -*)
                uerror "unrecognized option \`$arg'"
                ;;
            *)
                if test "x$target" = x
                then
                    target="$arg"
                else
                    dependencies="$dependencies $arg"
                fi
        esac
    else
        if test "x$target" = x
        then
            target="$arg"
        else
            dependencies="$dependencies $arg"
        fi
    fi
done
unset arg onlyargs

if test "x$target" = x
then
    uerror "too few arguments"
fi
# Sanity check that all required variables are provided.
info -n 3 "checking that all required environment variables are set... "
for v in $envvars
do
    if eval test '"x${'$v'+set}"' != xset
    then
        test $verbosity -lt 3 || echo "" >&2
        error "$v: required environment variable not set"
    fi
done
unset v
test $verbosity -lt 3 || echo "ok" >&2
info -n 3 "checking TEXLIMIT... "
if echo "$TEXLIMIT" | grep -q '^[0-9]*$'
then
    test $verbosity -lt 3 || echo "$TEXLIMIT" >&2
else
    test $verbosity -lt 3 || echo "" >&2
    error "TEXLIMIT: not an integer ($TEXLIMIT)"
fi

jobname="${target%.*}"
directory="$(${DIRNAME} "$target")"
abs_srcdir="$(readlink -f "$abs_srcdir")"
abs_builddir="$(readlink -f "$abs_builddir")"

## usage: printfgen TOOL FILE
##
## Report that TOOL (which should be a generic name like 'TEX') is to be run on
## FILE.
##
printfgen()
{
    printf '  %-9s%s\n' "$1" "$2"
}

## usage: soft_link SOURCE TARGET
##
## Create the soft link TARGET for file SOURCE creating directories as needed.
## If verbosity is high enough, output progress information.
##
soft_link()
{
    d="$(${DIRNAME} "$2")"
    if test $verbosity -gt 1; then printfgen MKDIR_P "$d"; fi
    ${MKDIR_P} "$d"
    if test $verbosity -gt 1; then printfgen LN_S "$2 --> $1"; fi
    ${LN_S} "$1" "$2"
    unset d
}

## usage: cat_log FILE
##
## Output the last 32 lines of log file FILE enclosed between horizontal bars
## and a note that prints the absolute file name of FILE.
##
cat_log()
{
    sb='------------'
    lf="$(readlink -f "$1")"
    echo "# $sb $lf $sb"
    tail -n 32 "$1"
    echo "$sb $lf $sb" | ${SED} -e 's,.,-,g' -e 's,^,# ,g'
    unset sb lf
}

## usage: src_to_build PATH
##
## Output an absolute file name that is obtained by replacing the trailing
## *abs_srcdir* component of PATH with *abs_builddir*.  PATH will be expanded
## to a canonical path if needed.
##
src_to_build()
{
    prefix_length=$(echo "$abs_srcdir" | ${AWK} '{ print length; }')
    p="$(readlink -f $1)"
    p=$(echo "$p" | ${SED} "s,^.\{$prefix_length\},,g")
    p="$abs_builddir/$p"
    echo "$p"
    unset prefix_length p
}

if test "$abs_srcdir" = "$abs_builddir"
then
    info 1 "no VPATH build"
else
    for src in $dependencies
    do
        src="$(readlink -f "$src")"
        bld="$(src_to_build "$src")"
        test -e "$bld" || soft_link "$src" "$bld" || exit 1
    done
    unset src bld
fi
info 2 "entering directory \`$directory'"
cd "$directory"
printfgen GEN "$directory/$target"
touch "$jobname.aux"
i=0
while :
do
    i=$(($i + 1))
    md5aux="$(md5sum *.aux)"
    printfgen TEX "$directory/$jobname[.tex]"
    if ${TEX} ${TEXFLAGS} "$jobname" > "$jobname.stdout"
    then
        rm -f "$jobname.stdout"
    else
        cat_log "$jobname.stdout"
        error "TEX ($(echo $TEX $TEXFLAGS))" \
              "encountered an operational error"
    fi
    if test $i -eq 1
    then
        # Some packages that add support for multiple bibliographies will
        # require BibTeX be run on more than one file.
        for a in *.aux
        do
            # Running 'bibtex jobname.aux' is the same as running
            # 'bibtex jobname' but Biber requires that you omit the '.aux'
            # suffix...
            auxbase="${a%.aux}"
            if grep -q -e '\citation{' -e '\abx@aux@cite{' "$a"
            then
                printfgen BIB "${directory}/${auxbase}[.aux]"
                if ${BIB} ${BIBFLAGS} "$auxbase" > "${auxbase}.stdout"
                then
                    rm -f "${auxbase}.stdout"
                else
                    cat_log "${auxbase}.stdout"
                    error "$a: BIB ($(echo $BIB $BIBFLAGS))" \
                          "encountered an operational error"
                fi
            fi
        done
        unset a
    elif echo "$md5aux" | md5sum --check 1> /dev/null 2> /dev/null
    then
        break
    elif test $i -ge ${TEXLIMIT}
    then
        error "TeX limit exceeded (TEXLIMIT = $TEXLIMIT)"
    fi
done
